import asyncio
from functools import wraps
from json import JSONDecodeError, loads
from typing import Callable, Dict, Literal, Tuple, TypeVar

from d42 import substitute
from jj import RelayResponse
from revolt.errors import SubstitutionError
from schemax_openapi import SchemaData
from valera import ValidationException

from ._config import Config
from .utils import create_openapi_matcher, load_cache

_T = TypeVar('_T')

class Validator:

    def __init__(self,
                 is_raise_error: bool,
                 func_name: str,
                 spec_link: str | None = None,
                 validate_level: Literal["error", "warning"] = "warning",
                 prefix: str | None = None,
                 ):
        self.func_name = func_name
        self.spec_link = spec_link
        self.is_raise_error = is_raise_error
        self.validate_level = validate_level
        self.prefix = prefix

    def output(self,
               e: Exception):
        if Config.OUTPUT_FUNCTION is None:
            print(f"⚠️ ⚠️ ⚠️ There are some mismatches in {self.func_name} :\n{str(e)}\n")
        else:
            Config.OUTPUT_FUNCTION(self, e)

    def _validation_failure(self,
                            exception: Exception,
                            ) -> None:
        self.output(exception)
        if self.is_raise_error:
            raise ValidationException(f"There are some mismatches in {self.func_name}:\n{str(exception)}")

    def prepare_data(self) -> Dict[Tuple[str, str], SchemaData]:
        return load_cache(self.spec_link, self.func_name)

    def _prepare_validation(self,
                           mocked: _T,
                           ):
        mock_matcher = mocked.handler.matcher

        try:
            # check for JSON in response of mock
            decoded_mocked_body = loads(mocked.handler.response.get_body().decode())  # type: ignore
        except JSONDecodeError:
            raise AssertionError(f"JSON expected in Response body of the {self.func_name}")

        spec_matcher = create_openapi_matcher(matcher=mock_matcher, prefix=self.prefix)

        if not spec_matcher:
            raise AssertionError(f"There is no valid matcher in {self.func_name}")

        prepared_spec = self.prepare_data()
        matched_spec_units = [(http_method, path) for http_method, path in prepared_spec.keys() if
                              spec_matcher.match((http_method, path))]

        if len(matched_spec_units) > 1:
            raise AssertionError(f"There is more than 1 matches")

        if len(matched_spec_units) == 0:
            raise AssertionError(f"API method '{prepared_spec}' was not found in the spec_link "
                                 f"for the validation of {self.func_name}")

        spec_unit = prepared_spec.get(matched_spec_units[0])

        return spec_unit, decoded_mocked_body


    def validate(self,
                 mocked: _T,
                 ) -> None:

        spec_unit, decoded_mocked_body = self._prepare_validation(mocked=mocked)

        if spec_unit.response_schema_d42:

            try:
                substitute(spec_unit.response_schema_d42, decoded_mocked_body)
            except SubstitutionError as exception:
                self._validation_failure(exception)

        else:
            raise AssertionError(f"API method '{spec_unit}' in the spec_link"
                                 f" lacks a response structure for the validation of {self.func_name}")


def validate_spec(*,
                  spec_link: str | None,
                  is_raise_error: bool = None,
                  prefix: str | None = None
                  ) -> Callable[[Callable[..., _T]], Callable[..., _T]]:
    """
    Validates the jj mock function with given specification lint.

    Args:
       spec_link: The link to the specification. `None` for disable validation.
       is_raise_error: If True - raises error when validation is failes. False is default.
       prefix: Prefix is used to cut paths prefix in mock function.
    """
    def decorator(func: Callable[..., _T]) -> Callable[..., _T]:
        func_name = func.__name__

        validator = Validator(
            spec_link=spec_link,
            prefix=prefix,
            func_name=func_name,
            is_raise_error=is_raise_error if is_raise_error is not None else Config.IS_RAISES
            )

        @wraps(func)
        async def async_wrapper(*args: object, **kwargs: object) -> _T:
            mocked = await func(*args, **kwargs)
            if validator.spec_link:
                if isinstance(mocked.handler.response, RelayResponse):
                    print("RelayResponse type is not supported")
                    return mocked
                validator.validate(mocked)
            else:...
            return mocked

        @wraps(func)
        def sync_wrapper(*args: object, **kwargs: object) -> _T:
            mocked = func(*args, **kwargs)
            if validator.spec_link:
                if isinstance(mocked.handler.response, RelayResponse):
                    print("RelayResponse type is not supported")
                    return mocked
                validator.validate(mocked)
            else:...
            return mocked

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper
    return decorator



import json
from hashlib import md5
from os import makedirs, path, remove
from pickle import dump
from pickle import load as pickle_load
from time import time
from typing import Dict, List, Tuple

from httpx import ConnectTimeout, Response, get
from schemax_openapi import SchemaData, collect_schema_data
from yaml import CLoader, load

from .._config import Config

__all__ = ('load_cache', )


CACHE_DIR = Config.MAIN_DIRECTORY + '/_cache_parsed_specs'
CACHE_TTL = 3600  # in second


def _build_entity_dict(entities: List[SchemaData]) -> Dict[Tuple[str, str], SchemaData]:
    entity_dict = {}
    for entity in entities:
        entity_key = (entity.http_method.upper(), entity.path)
        entity_dict[entity_key] = entity
    return entity_dict


def _validate_cache_file(filename: str) -> bool:
    if not path.isfile(filename):
        return False

    file_age = time() - path.getmtime(filename)

    if file_age > CACHE_TTL:
        remove(filename)
        return False

    return True


def _get_cache_filename(url: str) -> str:
    hash_obj = md5(url.encode())
    return path.join(CACHE_DIR, hash_obj.hexdigest() + '.cache' + '.yml')


def _download_spec(spec_link: str) -> Response:
    try:
        response = get(spec_link)
    except ConnectTimeout:
        raise ConnectTimeout(f"Timeout occurred while trying to connect to the {spec_link}.")
    response.raise_for_status()
    return response


def _save_cache(spec_link: str, raw_schema) -> None:
    filename = _get_cache_filename(spec_link)
    makedirs(CACHE_DIR, exist_ok=True)
    with open(filename, 'wb') as f:
        dump(raw_schema, f)


def load_cache(spec_link: str, func_name: str | None = None) -> Dict[Tuple[str, str], SchemaData]:
    filename = _get_cache_filename(spec_link)

    if _validate_cache_file(filename):
        with TimeitContext(f"LOAD parsed YAML of {spec_link} from cache", custom_str=f"{func_name}"):
            with open(filename, 'rb') as f:
                raw_schema = pickle_load(f)
    else:
        raw_spec = _download_spec(spec_link)

        content_type = raw_spec.headers.get('Content-Type', '')

        if 'application/json' in content_type:
            raw_schema = json.loads(raw_spec.text)
        elif 'text/yaml' in content_type or 'application/x-yaml' in content_type:
            raw_schema = load(raw_spec.text, Loader=CLoader)
        else:
            raise ValueError(f"Unsupported content type: {content_type}")

        _save_cache(spec_link, raw_schema)

    parsed_data = collect_schema_data(raw_schema)
    prepared_dict = _build_entity_dict(parsed_data)

    return prepared_dict



import re

__all__ = ('normalize_path', 'destroy_prefix', )


def normalize_path(path: str) -> str:
    var_counter = 1

    def replace_var(match) -> str:
        nonlocal var_counter
        replacement = f'{{var{var_counter}}}'
        var_counter += 1
        return replacement

    normalized_path = re.sub(r'{[a-zA-Z0-9_]+}', replace_var, path)

    return normalized_path


def destroy_prefix(path: str, prefix: str) -> str:
    return re.sub(prefix, '', path)
